#include "ADC.h"

volatile uint32_t adc_value;
volatile int32_t samp_store[N_STORE];
volatile int32_t hilbert_coeffs[N_HILBERT];

void init_ADC()
{
    // calc hilbert coefficients
    for(int32_t i = -((N_HILBERT-1)/2); i <= ((N_HILBERT-1)/2); i++)
    {
        if(((i + 1) % 2) == 1)
        {
            hilbert_coeffs[i + ((N_HILBERT-1)/2)] = (int32_t)(((double)2048) * (((double)2.0) / (((double)M_PI) * ((double)i))));
        }
        else
        {
            hilbert_coeffs[i + ((N_HILBERT-1)/2)] = 0;
        }
    }
    
    GPIOC->MODER &= ~(0x3 << 0);
    GPIOC->MODER |= (0x3 << 0);    // PC0 (ADC1_IN10) in analog mode
    
    RCC->APB2ENR |= (1 << 8) | (1 << 9) | (1 << 10);    // enable clock for ADC1, ADC2 and ADC3
    
    ADC1->SQR1 &= ~(0xF << 20);    // L = 0
    ADC1->SQR1 |= (0x1 << 20);     // L = 0
    ADC1->SQR3 &= ~(0x1F << 0);    // SQ1 = 0
    ADC1->SQR3 |= ((0x1F & 10) << 0);       // SQ1 = 10, so we use ADC1_IN10 (PC0)
    
    ADC1->SMPR2 |= (0x1 << 0);    // 15 cycles sample time (channel 0)
    ADC1->SMPR1 |= (0x1 << 0);    // 15 cycles sample time (channel 10)
    
    ADC1->CR2 |= (1 << 8);    // enable DMA
    ADC1->CR2 |= (1 << 9);    // enable DMA requests
    ADC1->CR2 |= (1 << 1);    // enable continous mode
    
    
    RCC->AHB1ENR |= (1 << 22);     // enable clock for DMA2 (which is used for ADC1, ADC2 and ADC3)
    NVIC->ISER[1] |= (1 << 28);    // enable interrupt 60 (DMA2_Stream4) (60 = 32 + 28)
    
    while(((DMA2->S4CR >> 0) & 0x1) == 1);
    
    DMA2->S4NDTR = 0xFFFF & 1;     // Transfer 1 data item in stream 4
    DMA2->S4PAR = (uint32_t)(&(ADC1->DR));     // origin address is ADC1's data register
    DMA2->S4M0AR = ((uint32_t)(&adc_value));     // destination address is the address of the variable adc_value
    
    DMA2->S4CR &= ~(0xF << 25);    // CHSEL = 0, to use channel 0 in stream 4 (ADC1)
    
    DMA2->HIFCR |= (1 << 5);       // CTCIF4 = 1 ==> reset TCIF4
    DMA2->S4CR |= (1 << 4);        // TCIE = 1 ==> transfer complete interrupt enabled
    DMA2->S4CR &= ~(0x3 << 11);
    DMA2->S4CR |= (0x2 << 11);     // peripheral data size = 32bit (word)
    
    DMA2->S4CR &= ~(0x3 << 13);
    DMA2->S4CR |= (0x2 << 13);     // memory size = 32bit (word)
    
    DMA2->S4CR |= (1 << 0);        // EN = 1 ==> stream 4 enabled
    
    
    ADC1->CR2 |= (1 << 0);    // enable ADC
    ADC1->CR2 |= (1 << 30);   // start conversion of regular channels
}

void DMA2_Stream4_IRQHandler()
{
    static int32_t I_old = 0;
    static int32_t Q_old = 0;
    
    if(((DMA2->HISR >> 5) & 0x1) == 1)    // TCIF4 = 1 ==> a transfer complete event happened on stream 4
    {
        DMA2->HIFCR |= (1 << 5);      // CTCIF4 = 1 ==> reset TCIF4
        DMA2->S4CR |= (1 << 0);       // EN = 1 ==> stream 4 enabled (why do we need to set this bit again in the interrupt handler? why and when is it reset?)
        
        GPIOA->ODR |= (1 << 8);
        
        // remove offset and adjust scaling
        int32_t adc_value_processed = (((int32_t)adc_value) - 1209) * 4;
        
        // push back and add new element
        for(uint32_t i = (N_STORE - 1); i >= 1 ; i--)
        {
            samp_store[i] = samp_store[i - 1];
        }
        samp_store[0] = adc_value_processed;
        
        // perform hilbert transformation by convolution
        int32_t Q = 0;
        int32_t n = N_HILBERT - 1;    // or 4?
        for(int32_t k = 0; k < N_HILBERT; k++)
        {
            if(((n - k) >= 0) && ((n - k) < N_HILBERT))
            {
                Q += (samp_store[k] * hilbert_coeffs[n - k]) / 1024;
            }
        }
        int32_t I = samp_store[4];
        
        // FM demodulation
        int32_t K = 512;
        int32_t diffangle = (    (K*Q*I_old) - (K*I*Q_old)    ) / ((I*I) + (Q*Q));
        I_old = I;
        Q_old = Q;
        
        int32_t L = 2048;
        if((diffangle > L) || (diffangle < (-L)))
        {
            diffangle = 0;
        }
        
        // output the sample using the DAC
        set_DAC1((uint32_t)(diffangle + ((int32_t)2048)));
        
        GPIOA->ODR &= ~(1 << 8);
    }
}
